package com.owlmaddie;

import com.mojang.blaze3d.systems.RenderSystem;
import net.fabricmc.api.ClientModInitializer;
import net.fabricmc.fabric.api.client.rendering.v1.WorldRenderContext;
import net.fabricmc.fabric.api.client.rendering.v1.WorldRenderEvents;
import net.minecraft.client.MinecraftClient;
import net.minecraft.client.font.TextRenderer;
import net.minecraft.client.font.TextRenderer.TextLayerType;
import net.minecraft.client.render.*;
import net.minecraft.client.util.math.MatrixStack;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityType;
import net.minecraft.entity.LivingEntity;
import net.minecraft.text.OrderedText;
import net.minecraft.text.StringVisitable;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.Box;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;
import org.joml.Matrix4f;
import org.joml.Quaternionf;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.util.List;
import java.util.Locale;
import java.util.stream.Collectors;


public class ClientInit implements ClientModInitializer {
    public static final Logger LOGGER = LoggerFactory.getLogger("mobgpt");
    protected static TextureLoader textures = new TextureLoader();;
    public static int DISPLAY_NUM_LINES = 3;
    public static int DISPLAY_PADDING = 2;

	@Override
    public void onInitializeClient() {
        ClickHandler.register();

        WorldRenderEvents.LAST.register((context) -> {
            drawTextAboveEntities(context, context.tickDelta());
        });
    }

    public void drawTextBubbleBackground(MatrixStack matrices, Entity entity, float x, float y, float width, float height) {
        RenderSystem.enableDepthTest();
        RenderSystem.enableBlend();
        RenderSystem.defaultBlendFunc();
        RenderSystem.setShader(GameRenderer::getPositionTexProgram);

        Tessellator tessellator = Tessellator.getInstance();
        BufferBuilder buffer = tessellator.getBuffer();
        float z = 0.01F;

        // Draw UI text background
        RenderSystem.setShaderTexture(0, textures.Get("ui", "text-top"));
        drawTexturePart(matrices, buffer, x, y, z, width, 40);

        RenderSystem.setShaderTexture(0, textures.Get("ui", "text-middle"));
        drawTexturePart(matrices, buffer, x, y + 40, z, width, height);

        RenderSystem.setShaderTexture(0, textures.Get("ui", "text-bottom"));
        drawTexturePart(matrices, buffer, x, y + 40 + height, z, width, 5);

        RenderSystem.disableBlend();
        RenderSystem.disableDepthTest();
    }

    private void drawTexturePart(MatrixStack matrices, BufferBuilder buffer, float x, float y, float z, float width, float height) {
        buffer.begin(VertexFormat.DrawMode.QUADS, VertexFormats.POSITION_TEXTURE);
        buffer.vertex(matrices.peek().getPositionMatrix(), x, y + height, z).texture(0, 1).next();  // bottom left
        buffer.vertex(matrices.peek().getPositionMatrix(), x + width, y + height, z).texture(1, 1).next();   // bottom right
        buffer.vertex(matrices.peek().getPositionMatrix(), x + width, y, z).texture(1, 0).next();  // top right
        buffer.vertex(matrices.peek().getPositionMatrix(), x, y, z).texture(0, 0).next(); // top left
        Tessellator.getInstance().draw();
    }

    private void drawStartIcon(MatrixStack matrices, Entity entity, float x, float y, float width, float height) {
        // Draw button icon
        Identifier button_texture = textures.Get("ui", "button-chat");
        RenderSystem.setShaderTexture(0, button_texture);
        RenderSystem.enableDepthTest();
        RenderSystem.enableBlend();
        RenderSystem.defaultBlendFunc();
        RenderSystem.setShader(GameRenderer::getPositionTexProgram);

        Tessellator tessellator = Tessellator.getInstance();
        BufferBuilder bufferBuilder = tessellator.getBuffer();
        bufferBuilder.begin(VertexFormat.DrawMode.QUADS, VertexFormats.POSITION_TEXTURE);

        float z = -0.01F;
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x, y + height, z).texture(0, 1).next();  // bottom left
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x + width, y + height, z).texture(1, 1).next();   // bottom right
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x + width, y, z).texture(1, 0).next();  // top right
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x, y, z).texture(0, 0).next(); // top left
        tessellator.draw();

        RenderSystem.disableBlend();
        RenderSystem.disableDepthTest();
    }

    private void drawEntityIcon(MatrixStack matrices, Entity entity, float x, float y, float width, float height) {
        // Draw face icon
        String entity_name = entity.getType().getUntranslatedName().toLowerCase(Locale.ROOT);
        Identifier entity_id = textures.Get("entity", entity_name);
        if (entity_id == null) {
            return;
        }

        RenderSystem.setShaderTexture(0, entity_id);
        RenderSystem.enableDepthTest();
        RenderSystem.enableBlend();
        RenderSystem.defaultBlendFunc();
        RenderSystem.setShader(GameRenderer::getPositionTexProgram);

        Tessellator tessellator = Tessellator.getInstance();
        BufferBuilder bufferBuilder = tessellator.getBuffer();
        bufferBuilder.begin(VertexFormat.DrawMode.QUADS, VertexFormats.POSITION_TEXTURE);

        float z = -0.01F;
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x, y + height, z).texture(0, 1).next();  // bottom left
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x + width, y + height, z).texture(1, 1).next();   // bottom right
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x + width, y, z).texture(1, 0).next();  // top right
        bufferBuilder.vertex(matrices.peek().getPositionMatrix(), x, y, z).texture(0, 0).next(); // top left
        tessellator.draw();

        RenderSystem.disableBlend();
        RenderSystem.disableDepthTest();
    }

    private void drawTextAboveEntities(WorldRenderContext context, float partialTicks) {
        Camera camera = context.camera();
        Entity cameraEntity = camera.getFocusedEntity();
        if (cameraEntity == null) return;

        World world = cameraEntity.getEntityWorld();
        double renderDistance = 9.0;

        // Calculate radius of entities
        Vec3d pos = cameraEntity.getPos();
        Box area = new Box(pos.x - renderDistance, pos.y - renderDistance, pos.z - renderDistance,
                           pos.x + renderDistance, pos.y + renderDistance, pos.z + renderDistance);

        // Init font render, matrix, and vertex producer
        TextRenderer fontRenderer = MinecraftClient.getInstance().textRenderer;
        MatrixStack matrices = context.matrixStack();
        VertexConsumerProvider immediate = context.consumers();

        // Get camera position
        Vec3d interpolatedCameraPos;
        if (MinecraftClient.getInstance().options.getPerspective().isFirstPerson()) {
            // 1st person, Use the cameraEntity (interpolate for smooth motion)
            interpolatedCameraPos = new Vec3d(
                    MathHelper.lerp(partialTicks, cameraEntity.prevX, cameraEntity.getPos().x),
                    MathHelper.lerp(partialTicks, cameraEntity.prevY, cameraEntity.getPos().y) + cameraEntity.getHeight(),
                    MathHelper.lerp(partialTicks, cameraEntity.prevZ, cameraEntity.getPos().z)
            );
        } else {
            // Use the camera (3rd person)
            interpolatedCameraPos = new Vec3d(camera.getPos().x,
                    camera.getPos().y,
                    camera.getPos().z);
        }

        // Get all entities
        List<Entity> nearbyEntities = world.getOtherEntities(null, area);

        // Filter living entities
        List<LivingEntity> nearbyCreatures = nearbyEntities.stream()
                .filter(entity -> entity instanceof LivingEntity)
                .map(entity -> (LivingEntity) entity)
                .collect(Collectors.toList());

        for (Entity entity : nearbyCreatures) {
            if (entity.getType() == EntityType.PLAYER) {
                // Skip Player
                continue;
            }

            // Look-up greeting (if any)
            ChatDataManager.EntityChatData chatData = ChatDataManager.getInstance().getOrCreateChatData(entity.getId());
            List<String> lines = chatData.getWrappedLines();

            // Set the range of lines to display
            int starting_line = chatData.currentLineNumber;
            int ending_line = Math.min(chatData.currentLineNumber + DISPLAY_NUM_LINES, lines.size());

            // Push a new matrix onto the stack.
            matrices.push();

            // Translate to the entity's position
            double paddingAboveEntity = 0.5D;
            matrices.translate(entity.getPos().x - interpolatedCameraPos.x,
                               entity.getPos().y - interpolatedCameraPos.y + entity.getHeight() + paddingAboveEntity,
                               entity.getPos().z - interpolatedCameraPos.z);

            // Calculate the difference vector (from entity to camera)
            Vec3d difference = cameraEntity.getPos().subtract(entity.getPos());

            // Calculate the yaw angle (just like before)
            double yaw = -(Math.atan2(difference.z, difference.x) + Math.PI / 2D);

            // Calculate the pitch difference (using y component)
            double pitch = Math.atan2(difference.y, Math.sqrt(difference.x * difference.x + difference.z * difference.z));

            // Clamp the pitch to the desired range (in this case, ±X degrees converted to radians)
            pitch = MathHelper.clamp(pitch, -Math.toRadians(20), Math.toRadians(20));

            // Convert yaw and pitch to Quaternionf
            float halfYaw = (float) yaw * 0.5f;
            double sinHalfYaw = MathHelper.sin(halfYaw);
            double cosHalfYaw = MathHelper.cos(halfYaw);

            float halfPitch = (float) pitch * 0.5f;
            double sinHalfPitch = MathHelper.sin(halfPitch);
            double cosHalfPitch = MathHelper.cos(halfPitch);

            // Constructing the Quaternionf for yaw (around Y axis) and pitch (around X axis)
            Quaternionf yawRotation = new Quaternionf(0, sinHalfYaw, 0, cosHalfYaw);
            Quaternionf pitchRotation = new Quaternionf(sinHalfPitch, 0, 0, cosHalfPitch);

            // Combine the rotations
            yawRotation.mul(pitchRotation);

            // Now when you want to render, apply the combined rotation:
            matrices.multiply(yawRotation);

            // Determine max line length
            float linesDisplayed = ending_line - starting_line;
            float lineSpacing = 1F;
            float textHeaderHeight = 40F;
            float textFooterHeight = 5F;


            // Calculate size of text scaled to world
            float scaledTextHeight = linesDisplayed * (fontRenderer.fontHeight + lineSpacing);
            float minTextHeight = (DISPLAY_NUM_LINES * (fontRenderer.fontHeight + lineSpacing)) + (DISPLAY_PADDING * 2);
            scaledTextHeight = Math.max(scaledTextHeight, minTextHeight);

            // Scale down before rendering textures (otherwise font is huge)
            matrices.scale(-0.02F, -0.02F, 0.02F);

            // Translate above the entity
            matrices.translate(0F, -scaledTextHeight + -textHeaderHeight + -textFooterHeight, 0F);

            // Check if conversation has started
            if (chatData.currentMessage.isEmpty()) {
                // Draw 'start' button
                drawStartIcon(matrices, entity, -16, textHeaderHeight, 32, 17);

            } else {
                // Draw text background (no smaller than 50F tall)
                drawTextBubbleBackground(matrices, entity, -64, 0, 128, scaledTextHeight);

                // Draw face of entity
                drawEntityIcon(matrices, entity, -59, 7, 32, 32);

                // Render each line of the text
                int fullBright = 0xF000F0;
                Matrix4f matrix = matrices.peek().getPositionMatrix();
                float yOffset = 40.0F + DISPLAY_PADDING;
                int currentLineIndex = 0; // We'll use this to track which line we're on

                for (String lineText : lines) {
                    // Only draw lines that are within the specified range
                    if (currentLineIndex >= starting_line && currentLineIndex < ending_line) {
                        fontRenderer.draw(lineText, -fontRenderer.getWidth(lineText) / 2f, yOffset, 0xffffff,
                                false, matrix, immediate, TextLayerType.NORMAL, 0, fullBright);
                        yOffset += fontRenderer.fontHeight + lineSpacing;
                    }
                    currentLineIndex++;

                    if (currentLineIndex > ending_line) {
                        break;
                    }
                }

                // Add end of message
                if (starting_line > 0 && starting_line == ending_line) {
                    String lineText = "<end of message>";
                    fontRenderer.draw(lineText, -fontRenderer.getWidth(lineText) / 2f, yOffset + 10F, 0xffffff,
                            false, matrix, immediate, TextLayerType.NORMAL, 0, fullBright);
                }
            }

            // Pop the matrix to return to the original state.
            matrices.pop();
        }
    }
    
}
